import typing

import keras
import numpy as np
from keras import backend
from keras import layers

from kimm.blocks import apply_conv2d_block
from kimm.blocks import apply_se_block
from kimm.models.base_model import BaseModel
from kimm.utils import add_model_to_registry


def _adjust_widths_and_groups(widths, groups, expansion_ratio):
    def _quantize_float(f, q):
        return int(round(f / q) * q)

    bottleneck_widths = [int(w * b) for w, b in zip(widths, expansion_ratio)]
    groups = [min(g, w_bot) for g, w_bot in zip(groups, bottleneck_widths)]
    bottleneck_widths = [
        _quantize_float(w_bot, g) for w_bot, g in zip(bottleneck_widths, groups)
    ]
    widths = [
        int(w_bot / b) for w_bot, b in zip(bottleneck_widths, expansion_ratio)
    ]
    return widths, groups


def _generate_regnet(
    width_init,
    width_slope,
    width_mult,
    group_size,
    depth,
    quant=8,
    expansion_ratio=1.0,
):
    widths_cont = np.arange(depth) * width_slope + width_init
    width_exps = np.round(np.log(widths_cont / width_init) / np.log(width_mult))
    widths = (
        np.round(
            np.divide(width_init * np.power(width_mult, width_exps), quant)
        )
        * quant
    )
    num_stages = len(np.unique(widths))
    groups = np.array([group_size for _ in range(num_stages)])

    widths = np.array(widths).astype(int).tolist()
    stage_gs = groups.astype(int).tolist()

    # Convert to per-stage format
    stage_widths, stage_depths = np.unique(widths, return_counts=True)
    stage_e = [expansion_ratio for _ in range(num_stages)]
    stage_strides = []
    for _ in range(num_stages):
        stride = 2
        stage_strides.append(stride)

    # Adjust the compatibility of ws and gws
    stage_widths, stage_gs = _adjust_widths_and_groups(
        stage_widths, stage_gs, stage_e
    )
    per_stage_args = [
        params
        for params in zip(
            stage_widths,
            stage_strides,
            stage_depths,
            stage_e,
            stage_gs,
        )
    ]
    return per_stage_args


def apply_bottleneck_block(
    inputs,
    output_channels: int,
    strides: int = 1,
    expansion_ratio: float = 1.0,
    group_size: int = 1,
    se_ratio: float = 0.25,
    activation="relu",
    linear_out: bool = False,
    name="bottleneck_block",
):
    channels_axis = -1 if backend.image_data_format() == "channels_last" else -3
    input_channels = inputs.shape[channels_axis]
    expansion_channels = int(round(output_channels * expansion_ratio))
    groups = expansion_channels // group_size

    shortcut = inputs
    x = inputs
    x = apply_conv2d_block(
        x,
        expansion_channels,
        1,
        1,
        activation=activation,
        name=f"{name}_conv1",
    )
    x = apply_conv2d_block(
        x,
        expansion_channels,
        3,
        strides,
        groups=groups,
        activation=activation,
        name=f"{name}_conv2",
    )
    if se_ratio > 0.0:
        x = apply_se_block(
            x,
            se_ratio,
            activation,
            se_input_channels=input_channels,
            name=f"{name}_se",
        )
    x = apply_conv2d_block(
        x,
        output_channels,
        1,
        1,
        activation=None,
        name=f"{name}_conv3",
    )

    # downsampling
    if strides != 1 or input_channels != output_channels:
        shortcut = apply_conv2d_block(
            shortcut,
            output_channels,
            1,
            strides,
            activation=None,
            name=f"{name}_downsample",
        )

    x = layers.Add(name=f"{name}_add")([x, shortcut])
    if not linear_out:
        x = layers.Activation(activation=activation, name=f"{name}")(x)
    return x


@keras.saving.register_keras_serializable(package="kimm")
class RegNet(BaseModel):
    available_feature_keys = [
        "STEM_S2",
        *[f"BLOCK{i}_S{j}" for i, j in zip(range(4), [4, 8, 16, 32])],
    ]

    def __init__(
        self,
        w0: int = 80,
        wa: float = 42.64,
        wm: float = 2.66,
        group_size: int = 24,
        depth: int = 21,
        se_ratio: float = 0.0,
        **kwargs,
    ):
        kwargs["weights_url"] = self.get_weights_url(kwargs["weights"])

        per_stage_config = _generate_regnet(w0, wa, wm, group_size, depth)

        input_tensor = kwargs.pop("input_tensor", None)
        self.set_properties(kwargs)
        inputs = self.determine_input_tensor(
            input_tensor,
            self._input_shape,
            self._default_size,
        )
        x = inputs

        x = self.build_preprocessing(x, "imagenet")

        # Prepare feature extraction
        features = {}

        # stem
        stem_channels = 32
        x = apply_conv2d_block(
            x, stem_channels, 3, 2, activation="relu", name="stem"
        )
        features["STEM_S2"] = x

        # stages
        current_stride = 2
        for current_stage_idx, params in enumerate(per_stage_config):
            c, s, d, e, g = params
            current_stride *= s
            # blocks
            for current_block_idx in range(d):
                s = s if current_block_idx == 0 else 1
                name = f"s{current_stage_idx + 1}_b{current_block_idx + 1}"
                x = apply_bottleneck_block(x, c, s, e, g, se_ratio, name=name)
            # add feature
            features[f"BLOCK{current_stage_idx}_S{current_stride}"] = x

        # Head
        x = self.build_head(x)

        super().__init__(inputs=inputs, outputs=x, features=features, **kwargs)

        # All references to `self` below this line
        self.w0 = w0
        self.wa = wa
        self.wm = wm
        self.group_size = group_size
        self.depth = depth
        self.se_ratio = se_ratio

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "w0": self.w0,
                "wa": self.wa,
                "wm": self.wm,
                "group_size": self.group_size,
                "depth": self.depth,
                "se_ratio": self.se_ratio,
            }
        )
        return config

    def fix_config(self, config):
        unused_kwargs = ["w0", "wa", "wm", "group_size", "depth", "se_ratio"]
        for k in unused_kwargs:
            config.pop(k, None)
        return config


"""
Model Definition
"""


class RegNetX002(RegNet):
    available_weights = [
        (
            "imagenet",
            RegNet.default_origin,
            "regnetx002_regnetx_002.pycls_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "RegNetX002",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            24,
            36.44,
            2.49,
            8,
            13,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class RegNetY002(RegNet):
    available_weights = [
        (
            "imagenet",
            RegNet.default_origin,
            "regnety002_regnety_002.pycls_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "RegNetY002",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            24,
            36.44,
            2.49,
            8,
            13,
            0.25,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class RegNetX004(RegNet):
    available_weights = [
        (
            "imagenet",
            RegNet.default_origin,
            "regnetx004_regnetx_004.pycls_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "RegNetX004",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            24,
            24.48,
            2.54,
            16,
            22,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class RegNetY004(RegNet):
    available_weights = [
        (
            "imagenet",
            RegNet.default_origin,
            "regnety004_regnety_004.tv2_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "RegNetY004",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            48,
            27.89,
            2.09,
            8,
            16,
            0.25,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class RegNetX006(RegNet):
    available_weights = [
        (
            "imagenet",
            RegNet.default_origin,
            "regnetx006_regnetx_006.pycls_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "RegNetX006",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            48,
            36.97,
            2.24,
            24,
            16,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class RegNetY006(RegNet):
    available_weights = [
        (
            "imagenet",
            RegNet.default_origin,
            "regnety006_regnety_006.pycls_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "RegNetY006",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            48,
            32.54,
            2.32,
            16,
            15,
            0.25,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class RegNetX008(RegNet):
    available_weights = [
        (
            "imagenet",
            RegNet.default_origin,
            "regnetx008_regnetx_008.tv2_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "RegNetX008",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            56,
            35.73,
            2.28,
            16,
            16,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class RegNetY008(RegNet):
    available_weights = [
        (
            "imagenet",
            RegNet.default_origin,
            "regnety008_regnety_008.pycls_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "RegNetY008",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            56,
            38.84,
            2.4,
            16,
            14,
            0.25,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class RegNetX016(RegNet):
    available_weights = [
        (
            "imagenet",
            RegNet.default_origin,
            "regnetx016_regnetx_016.tv2_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "RegNetX016",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            80,
            34.01,
            2.25,
            24,
            18,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class RegNetY016(RegNet):
    available_weights = [
        (
            "imagenet",
            RegNet.default_origin,
            "regnety016_regnety_016.tv2_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "RegNetY016",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            48,
            20.71,
            2.65,
            24,
            27,
            0.25,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class RegNetX032(RegNet):
    available_weights = [
        (
            "imagenet",
            RegNet.default_origin,
            "regnetx032_regnetx_032.tv2_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "RegNetX032",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            88,
            26.31,
            2.25,
            48,
            25,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class RegNetY032(RegNet):
    available_weights = [
        (
            "imagenet",
            RegNet.default_origin,
            "regnety032_regnety_032.ra_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "RegNetY032",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            80,
            42.63,
            2.66,
            24,
            21,
            0.25,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class RegNetX040(RegNet):
    available_weights = [
        (
            "imagenet",
            RegNet.default_origin,
            "regnetx040_regnetx_040.pycls_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "RegNetX040",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            96,
            38.65,
            2.43,
            40,
            23,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class RegNetY040(RegNet):
    available_weights = [
        (
            "imagenet",
            RegNet.default_origin,
            "regnety040_regnety_040.ra3_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "RegNetY040",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            96,
            31.41,
            2.24,
            64,
            22,
            0.25,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class RegNetX064(RegNet):
    available_weights = [
        (
            "imagenet",
            RegNet.default_origin,
            "regnetx064_regnetx_064.pycls_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "RegNetX064",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            184,
            60.83,
            2.07,
            56,
            17,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class RegNetY064(RegNet):
    available_weights = [
        (
            "imagenet",
            RegNet.default_origin,
            "regnety064_regnety_064.ra3_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "RegNetY064",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            112,
            33.22,
            2.27,
            72,
            25,
            0.25,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class RegNetX080(RegNet):
    available_weights = [
        (
            "imagenet",
            RegNet.default_origin,
            "regnetx080_regnetx_080.tv2_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "RegNetX080",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            80,
            49.56,
            2.88,
            120,
            23,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class RegNetY080(RegNet):
    available_weights = [
        (
            "imagenet",
            RegNet.default_origin,
            "regnety080_regnety_080.ra3_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "RegNetY080",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            192,
            76.82,
            2.19,
            56,
            17,
            0.25,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class RegNetX120(RegNet):
    available_weights = [
        (
            "imagenet",
            RegNet.default_origin,
            "regnetx120_regnetx_120.pycls_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "RegNetX120",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            168,
            73.36,
            2.37,
            112,
            19,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class RegNetY120(RegNet):
    available_weights = [
        (
            "imagenet",
            RegNet.default_origin,
            "regnety120_regnety_120.sw_in12k_ft_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "RegNetY120",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            168,
            73.36,
            2.37,
            112,
            19,
            0.25,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class RegNetX160(RegNet):
    available_weights = [
        (
            "imagenet",
            RegNet.default_origin,
            "regnetx160_regnetx_160.tv2_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "RegNetX160",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            216,
            55.59,
            2.1,
            128,
            22,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class RegNetY160(RegNet):
    available_weights = [
        (
            "imagenet",
            RegNet.default_origin,
            "regnety160_regnety_160.swag_ft_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "RegNetY160",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            200,
            106.23,
            2.48,
            112,
            18,
            0.25,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class RegNetX320(RegNet):
    available_weights = [
        (
            "imagenet",
            RegNet.default_origin,
            "regnetx320_regnetx_320.tv2_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "RegNetX320",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            320,
            69.86,
            2.0,
            168,
            23,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class RegNetY320(RegNet):
    available_weights = [
        (
            "imagenet",
            RegNet.default_origin,
            "regnety320_regnety_320.swag_ft_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "RegNetY320",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            232,
            115.89,
            2.53,
            232,
            20,
            0.25,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


add_model_to_registry(RegNetX002, "imagenet")
add_model_to_registry(RegNetY002, "imagenet")
add_model_to_registry(RegNetX004, "imagenet")
add_model_to_registry(RegNetY004, "imagenet")
add_model_to_registry(RegNetX006, "imagenet")
add_model_to_registry(RegNetY006, "imagenet")
add_model_to_registry(RegNetX008, "imagenet")
add_model_to_registry(RegNetY008, "imagenet")
add_model_to_registry(RegNetX016, "imagenet")
add_model_to_registry(RegNetY016, "imagenet")
add_model_to_registry(RegNetX032, "imagenet")
add_model_to_registry(RegNetY032, "imagenet")
add_model_to_registry(RegNetX040, "imagenet")
add_model_to_registry(RegNetY040, "imagenet")
add_model_to_registry(RegNetX064, "imagenet")
add_model_to_registry(RegNetY064, "imagenet")
add_model_to_registry(RegNetX080, "imagenet")
add_model_to_registry(RegNetY080, "imagenet")
add_model_to_registry(RegNetX120, "imagenet")
add_model_to_registry(RegNetY120, "imagenet")
add_model_to_registry(RegNetX160, "imagenet")
add_model_to_registry(RegNetY160, "imagenet")
add_model_to_registry(RegNetX320, "imagenet")
add_model_to_registry(RegNetY320, "imagenet")



import math
import typing

import keras
from keras import backend
from keras import layers
from keras import ops

from kimm.blocks import apply_conv2d_block
from kimm.blocks import apply_inverted_residual_block
from kimm.blocks import apply_transformer_block
from kimm.models.base_model import BaseModel
from kimm.utils import add_model_to_registry
from kimm.utils import make_divisible

# type, repeat, channels, strides, expansion_ratio, transformer_dim,
# transformer_depth, patch_size
DEFAULT_V1_S_CONFIG = [
    ["ir", 1, 3, 32, 1, 4.0, None, None, None],
    ["ir", 3, 3, 64, 2, 4.0, None, None, None],
    ["mobilevit", 1, 3, 96, 2, 4.0, 144, 2, 2],
    ["mobilevit", 1, 3, 128, 2, 4.0, 192, 4, 2],
    ["mobilevit", 1, 3, 160, 2, 4.0, 240, 3, 2],
]
DEFAULT_V1_XS_CONFIG = [
    ["ir", 1, 3, 32, 1, 4.0, None, None, None],
    ["ir", 3, 3, 48, 2, 4.0, None, None, None],
    ["mobilevit", 1, 3, 64, 2, 4.0, 96, 2, 2],
    ["mobilevit", 1, 3, 80, 2, 4.0, 120, 4, 2],
    ["mobilevit", 1, 3, 96, 2, 4.0, 144, 3, 2],
]
DEFAULT_V1_XXS_CONFIG = [
    ["ir", 1, 3, 16, 1, 2.0, None, None, None],
    ["ir", 3, 3, 24, 2, 2.0, None, None, None],
    ["mobilevit", 1, 3, 48, 2, 2.0, 64, 2, 2],
    ["mobilevit", 1, 3, 64, 2, 2.0, 80, 4, 2],
    ["mobilevit", 1, 3, 80, 2, 2.0, 96, 3, 2],
]


def unfold(inputs, patch_size):
    # TODO: improve performance
    x = inputs
    h, w, c = x.shape[-3], x.shape[-2], x.shape[-1]
    new_h, new_w = (
        math.ceil(h / patch_size) * patch_size,
        math.ceil(w / patch_size) * patch_size,
    )
    num_patches_h = new_h // patch_size
    num_patches_w = new_w // patch_size
    num_patches = num_patches_h * num_patches_w
    # [B, H, W, C] -> [B * P, N, C]
    x = ops.reshape(
        x, [-1, num_patches_h, patch_size, num_patches_w, patch_size, c]
    )
    x = ops.transpose(x, [0, 2, 4, 1, 3, 5])
    x = ops.reshape(x, [-1, num_patches, c])
    return x


def fold(inputs, h, w, c, patch_size):
    # TODO: improve performance
    x = inputs
    new_h, new_w = (
        math.ceil(h / patch_size) * patch_size,
        math.ceil(w / patch_size) * patch_size,
    )
    num_patches_h = new_h // patch_size
    num_patches_w = new_w // patch_size
    # [B * P, N, C] -> [B, P, N, C] -> [B, H, W, C]
    x = ops.reshape(
        x, [-1, patch_size, patch_size, num_patches_h, num_patches_w, c]
    )
    x = ops.transpose(x, [0, 3, 1, 4, 2, 5])
    x = ops.reshape(
        x, [-1, num_patches_h * patch_size, num_patches_w * patch_size, c]
    )
    return x


def apply_mobilevit_block(
    inputs,
    output_channels: int,
    kernel_size: int = 3,
    strides: int = 1,
    expansion_ratio: float = 1.0,
    mlp_ratio: float = 2.0,
    transformer_dim: typing.Optional[int] = None,
    transformer_depth: int = 2,
    patch_size: int = 8,
    num_heads: int = 4,
    projection_dropout_rate=0.0,
    attention_dropout_rate=0.0,
    activation="swish",
    transformer_activation="swish",
    fusion: bool = True,
    name="mobilevit_block",
):
    channels_axis = -1 if backend.image_data_format() == "channels_last" else -3
    input_channels = inputs.shape[channels_axis]
    transformer_dim = transformer_dim or make_divisible(
        input_channels * expansion_ratio
    )

    x = inputs

    # Local representation
    x = apply_conv2d_block(
        x,
        input_channels,
        kernel_size,
        strides,
        activation=activation,
        name=f"{name}_conv_kxk",
    )
    x = layers.Conv2D(
        transformer_dim, 1, use_bias=False, name=f"{name}_conv_1x1"
    )(x)

    # TODO: natively support channels_first
    # Unfold (feature map -> patches)
    if backend.image_data_format() == "channels_first":
        x = ops.transpose(x, [0, 2, 3, 1])

    h, w, c = x.shape[-3], x.shape[-2], x.shape[-1]
    x = unfold(x, patch_size)

    # Global representations
    for i in range(transformer_depth):
        x = apply_transformer_block(
            x,
            transformer_dim,
            num_heads,
            mlp_ratio,
            True,
            False,
            attention_dropout_rate=attention_dropout_rate,
            projection_dropout_rate=projection_dropout_rate,
            activation=transformer_activation,
            name=f"{name}_transformer_{i}",
        )
    x = layers.LayerNormalization(axis=-1, epsilon=1e-6, name=f"{name}_norm")(x)

    # Fold (patch -> feature map)
    x = fold(x, h, w, c, patch_size)

    # TODO: natively support channels_first
    if backend.image_data_format() == "channels_first":
        x = ops.transpose(x, [0, 3, 1, 2])

    x = apply_conv2d_block(
        x,
        output_channels,
        1,
        1,
        activation=activation,
        name=f"{name}_conv_proj",
    )
    if fusion:
        x = layers.Concatenate(axis=channels_axis)([inputs, x])

    x = apply_conv2d_block(
        x,
        output_channels,
        kernel_size,
        1,
        activation=activation,
        name=f"{name}_conv_fusion",
    )
    return x


@keras.saving.register_keras_serializable(package="kimm")
class MobileViT(BaseModel):
    available_feature_keys = [
        "STEM_S2",
        *[f"BLOCK{i}_S{j}" for i, j in zip(range(5), [2, 4, 8, 16, 32])],
    ]

    def __init__(
        self,
        stem_channels: int = 16,
        head_channels: int = 640,
        activation: str = "swish",
        config: str = "v1_s",
        **kwargs,
    ):
        kwargs["weights_url"] = self.get_weights_url(kwargs["weights"])

        _available_configs = ["v1_s", "v1_xs", "v1_xss"]
        if config == "v1_s":
            _config = DEFAULT_V1_S_CONFIG
        elif config == "v1_xs":
            _config = DEFAULT_V1_XS_CONFIG
        elif config == "v1_xxs":
            _config = DEFAULT_V1_XXS_CONFIG
        else:
            raise ValueError(
                f"config must be one of {_available_configs} using string. "
                f"Received: config={config}"
            )

        input_tensor = kwargs.pop("input_tensor", None)
        self.set_properties(kwargs, 256)

        inputs = self.determine_input_tensor(
            input_tensor,
            self._input_shape,
            self._default_size,
            static_shape=True,
        )
        x = inputs

        x = self.build_preprocessing(x, "0_1")

        # Prepare feature extraction
        features = {}

        # stem
        x = apply_conv2d_block(
            x, stem_channels, 3, 2, activation=activation, name="stem"
        )
        features["STEM_S2"] = x

        # blocks
        current_stride = 2
        for current_block_idx, cfg in enumerate(_config):
            (
                block_type,
                r,
                k,
                c,
                s,
                e,
                transformer_dim,
                transformer_depth,
                patch_size,
            ) = cfg
            # always apply inverted_residual_block
            for current_layer_idx in range(r):
                s = s if current_layer_idx == 0 else 1
                name = f"stages_{current_block_idx}_{current_layer_idx}"
                x = apply_inverted_residual_block(
                    x, c, k, 1, 1, s, e, activation=activation, name=name
                )
                current_stride *= s
            if block_type == "mobilevit":
                name = f"stages_{current_block_idx}_{current_layer_idx + 1}"
                x = apply_mobilevit_block(
                    x,
                    c,
                    k,
                    1,
                    transformer_dim=transformer_dim,
                    transformer_depth=transformer_depth,
                    patch_size=patch_size,
                    activation=activation,
                    transformer_activation=activation,
                    name=name,
                )
            features[f"BLOCK{current_block_idx}_S{current_stride}"] = x

        # last conv block
        x = apply_conv2d_block(
            x, head_channels, 1, 1, activation=activation, name="final_conv"
        )

        # Head
        x = self.build_head(x)

        super().__init__(inputs=inputs, outputs=x, features=features, **kwargs)

        # All references to `self` below this line
        self.stem_channels = stem_channels
        self.head_channels = head_channels
        self.activation = activation
        self.config = config

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "stem_channels": self.stem_channels,
                "head_channels": self.head_channels,
                "activation": self.activation,
                "config": self.config,
            }
        )
        return config

    def fix_config(self, config):
        unused_kwargs = [
            "stem_channels",
            "head_channels",
            "activation",
            "config",
        ]
        for k in unused_kwargs:
            config.pop(k, None)
        return config


class MobileViTS(MobileViT):
    available_weights = [
        (
            "imagenet",
            MobileViT.default_origin,
            "mobilevits_mobilevit_s.cvnets_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.1,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        config: str = "v1_s",
        name="MobileViTS",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            16,
            640,
            "swish",
            config,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class MobileViTXS(MobileViT):
    available_weights = [
        (
            "imagenet",
            MobileViT.default_origin,
            "mobilevitxs_mobilevit_xs.cvnets_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.1,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        config: str = "v1_xs",
        name="MobileViTXS",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            16,
            384,
            "swish",
            config,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class MobileViTXXS(MobileViT):
    available_weights = [
        (
            "imagenet",
            MobileViT.default_origin,
            "mobilevitxxs_mobilevit_xxs.cvnets_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.1,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        config: str = "v1_xxs",
        name="MobileViTXXS",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            16,
            320,
            "swish",
            config,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


add_model_to_registry(MobileViTS, "imagenet")
add_model_to_registry(MobileViTXS, "imagenet")
add_model_to_registry(MobileViTXXS, "imagenet")



import pathlib
import tempfile
import typing

from keras import backend
from keras import layers
from keras import models
from keras.src.utils.module_utils import tensorflow as tf

from kimm.models import BaseModel


def export_tflite(
    model: BaseModel,
    input_shape: typing.Union[int, typing.Sequence[int]],
    export_path: typing.Union[str, pathlib.Path],
    export_dtype: typing.Literal["float32", "float16", "int8"] = "float32",
    representative_dataset: typing.Optional[typing.Iterator] = None,
    batch_size: int = 1,
):
    """Export the model to tflite format.

    Only tensorflow backend with 'channels_last' is supported. The tflite model
    will be generated using `tf.lite.TFLiteConverter.from_saved_model` and
    optimized through tflite built-in functions.

    Note that when exporting an `int8` tflite model, `representative_dataset`
    must be passed.

    Args:
        model: keras.Model, the model to be exported.
        input_shape: int or sequence of int, specifying the shape of the input.
        export_path: str or pathlib.Path, specifying the path to export.
        export_dtype: str, specifying the export dtype.
        representative_dataset: None or Iterator, the calibration dataset for
            exporting int8 tflite.
        batch_size: int, specifying the batch size of the input,
            defaults to `1`.
    """
    if backend.backend() != "tensorflow":
        raise ValueError("`export_tflite` only supports tensorflow backend")
    if backend.image_data_format() != "channels_last":
        raise ValueError(
            "`export_tflite` only supports 'channels_last' data format."
        )
    if export_dtype not in ("float32", "float16", "int8"):
        raise ValueError(
            "`export_dtype` must be one of ('float32', 'float16', 'int8'). "
            f"Received: export_dtype={export_dtype}"
        )
    if export_dtype == "int8" and representative_dataset is None:
        raise ValueError(
            "For full integer quantization, a `representative_dataset` should "
            "be specified."
        )
    if isinstance(input_shape, int):
        input_shape = [input_shape, input_shape, 3]
    elif len(input_shape) == 2:
        input_shape = [input_shape[0], input_shape[1], 3]
    elif len(input_shape) == 3:
        input_shape = input_shape

    # Fix input shape
    inputs = layers.Input(shape=input_shape, batch_size=batch_size)
    outputs = model(inputs, training=False)
    model = models.Model(inputs, outputs)

    # Construct TFLiteConverter
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_path = pathlib.Path(temp_dir, "temp_saved_model")
        model.export(temp_path)
        converter = tf.lite.TFLiteConverter.from_saved_model(str(temp_path))

        # Configure converter
        if export_dtype != "float32":
            converter.optimizations = [tf.lite.Optimize.DEFAULT]
        if export_dtype == "int8":
            converter.target_spec.supported_ops = [
                tf.lite.OpsSet.TFLITE_BUILTINS_INT8
            ]
            converter.inference_input_type = tf.int8
            converter.inference_output_type = tf.int8
        elif export_dtype == "float16":
            converter.target_spec.supported_types = [tf.float16]
        if representative_dataset is not None:
            converter.representative_dataset = representative_dataset

        # Convert
        tflite_model = converter.convert()

    # Export
    with open(export_path, "wb") as f:
        f.write(tflite_model)



import pytest
from absl.testing import parameterized
from keras import layers
from keras.src import testing

from kimm.layers.position_embedding import PositionEmbedding


class PositionEmbeddingTest(testing.TestCase, parameterized.TestCase):
    @pytest.mark.requires_trainable_backend
    def test_position_embedding_basic(self):
        self.run_layer_test(
            PositionEmbedding,
            init_kwargs={},
            input_shape=(1, 10, 10),
            expected_output_shape=(1, 11, 10),
            expected_num_trainable_weights=2,
            expected_num_non_trainable_weights=0,
            expected_num_losses=0,
            supports_masking=False,
        )

    @pytest.mark.requires_trainable_backend
    def test_position_embedding_invalid_input_shape(self):
        inputs = layers.Input([3])
        with self.assertRaisesRegex(
            ValueError, "PositionEmbedding only accepts 3-dimensional input."
        ):
            PositionEmbedding()(inputs)



import pytest
from absl.testing import parameterized
from keras.src import testing

from kimm.layers.attention import Attention


class AttentionTest(testing.TestCase, parameterized.TestCase):
    @pytest.mark.requires_trainable_backend
    def test_attention_basic(self):
        self.run_layer_test(
            Attention,
            init_kwargs={"hidden_dim": 20, "num_heads": 2},
            input_shape=(1, 10, 20),
            expected_output_shape=(1, 10, 20),
            expected_num_trainable_weights=3,
            expected_num_non_trainable_weights=0,
            expected_num_losses=0,
            supports_masking=False,
        )



from kimm.utils.make_divisble import make_divisible
from kimm.utils.model_registry import add_model_to_registry
from kimm.utils.timm_utils import assign_weights
from kimm.utils.timm_utils import is_same_weights
from kimm.utils.timm_utils import separate_torch_state_dict



import typing

import keras
from keras import backend
from keras import layers

from kimm.blocks import apply_conv2d_block
from kimm.models import BaseModel
from kimm.utils import add_model_to_registry


def apply_dense_layer(
    inputs, growth_rate, expansion_ratio=4.0, name="dense_layer"
):
    channels_axis = -1 if backend.image_data_format() == "channels_last" else -3

    x = inputs
    x = layers.BatchNormalization(
        axis=channels_axis, momentum=0.9, epsilon=1e-5, name=f"{name}_norm1"
    )(x)
    x = layers.ReLU()(x)
    x = apply_conv2d_block(
        x,
        int(growth_rate * expansion_ratio),
        1,
        1,
        activation="relu",
        name=f"{name}_conv1",
    )
    x = layers.Conv2D(
        growth_rate, 3, 1, padding="same", use_bias=False, name=f"{name}_conv2"
    )(x)
    return x


def apply_dense_block(
    inputs, num_layers, growth_rate, expansion_ratio=4.0, name="dense_block"
):
    channels_axis = -1 if backend.image_data_format() == "channels_last" else -3

    x = inputs
    features = [x]
    for i in range(num_layers):
        new_features = layers.Concatenate(axis=channels_axis)(features)
        new_features = apply_dense_layer(
            new_features,
            growth_rate,
            expansion_ratio,
            name=f"{name}_denselayer{i + 1}",
        )
        features.append(new_features)
    x = layers.Concatenate(axis=channels_axis)(features)
    return x


def apply_dense_transition_block(
    inputs, output_channels, name="dense_transition_block"
):
    channels_axis = -1 if backend.image_data_format() == "channels_last" else -3
    x = inputs
    x = layers.BatchNormalization(
        axis=channels_axis, momentum=0.9, epsilon=1e-5, name=f"{name}_norm"
    )(x)
    x = layers.ReLU()(x)
    x = layers.Conv2D(
        output_channels, 1, 1, "same", use_bias=False, name=f"{name}_conv"
    )(x)
    x = layers.AveragePooling2D(
        2, 2, data_format=backend.image_data_format(), name=f"{name}_pool"
    )(x)
    return x


@keras.saving.register_keras_serializable(package="kimm")
class DenseNet(BaseModel):
    available_feature_keys = [
        "STEM_S4",
        *[f"BLOCK{i}_S{j}" for i, j in zip(range(4), [8, 16, 32, 32])],
    ]

    def __init__(
        self,
        growth_rate: float = 32,
        num_blocks: typing.Sequence[int] = [6, 12, 24, 16],
        **kwargs,
    ):
        kwargs["weights_url"] = self.get_weights_url(kwargs["weights"])
        input_tensor = kwargs.pop("input_tensor", None)
        self.set_properties(kwargs)
        channels_axis = (
            -1 if backend.image_data_format() == "channels_last" else -3
        )

        inputs = self.determine_input_tensor(
            input_tensor,
            self._input_shape,
            self._default_size,
        )
        x = inputs

        x = self.build_preprocessing(x, "imagenet")

        # Prepare feature extraction
        features = {}

        # Stem block
        stem_channel = growth_rate * 2
        x = apply_conv2d_block(
            x, stem_channel, 7, 2, activation="relu", name="features_conv0"
        )
        x = layers.ZeroPadding2D(1, name="features_pad0")(x)
        x = layers.MaxPooling2D(3, 2, name="features_pool0")(x)
        features["STEM_S4"] = x

        # Blocks
        current_stride = 4
        input_channels = stem_channel
        for current_block_idx, num_layers in enumerate(num_blocks):
            x = apply_dense_block(
                x,
                num_layers,
                growth_rate,
                expansion_ratio=4.0,
                name=f"features_denseblock{current_block_idx + 1}",
            )
            input_channels = input_channels + num_layers * growth_rate
            if current_block_idx != len(num_blocks) - 1:
                current_stride *= 2
                x = apply_dense_transition_block(
                    x,
                    input_channels // 2,
                    name=f"features_transition{current_block_idx + 1}",
                )
                input_channels = input_channels // 2

            features[f"BLOCK{current_block_idx}_S{current_stride}"] = x

        # Final batch norm
        x = layers.BatchNormalization(
            axis=channels_axis,
            momentum=0.9,
            epsilon=1e-5,
            name="features_norm5",
        )(x)
        x = layers.ReLU()(x)

        # Head
        x = self.build_head(x)

        super().__init__(inputs=inputs, outputs=x, features=features, **kwargs)

        # All references to `self` below this line
        self.growth_rate = growth_rate
        self.num_blocks = num_blocks

    def get_config(self):
        config = super().get_config()
        config.update(
            {"growth_rate": self.growth_rate, "num_blocks": self.num_blocks}
        )
        return config

    def fix_config(self, config: typing.Dict):
        unused_kwargs = ["growth_rate", "num_blocks"]
        for k in unused_kwargs:
            config.pop(k, None)
        return config


"""
Model Definition
"""


class DenseNet121(DenseNet):
    available_weights = [
        (
            "imagenet",
            DenseNet.default_origin,
            "densenet121_densenet121.ra_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "DenseNet121",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            32,
            [6, 12, 24, 16],
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            default_size=288,
            **kwargs,
        )


class DenseNet161(DenseNet):
    available_weights = [
        (
            "imagenet",
            DenseNet.default_origin,
            "densenet161_densenet161.tv_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "DenseNet161",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            48,
            [6, 12, 36, 24],
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            default_size=224,
            **kwargs,
        )


class DenseNet169(DenseNet):
    available_weights = [
        (
            "imagenet",
            DenseNet.default_origin,
            "densenet169_densenet169.tv_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "DenseNet169",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            32,
            [6, 12, 32, 32],
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            default_size=224,
            **kwargs,
        )


class DenseNet201(DenseNet):
    available_weights = [
        (
            "imagenet",
            DenseNet.default_origin,
            "densenet201_densenet201.tv_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "DenseNet201",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            32,
            [6, 12, 48, 32],
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            default_size=224,
            **kwargs,
        )


add_model_to_registry(DenseNet121, "imagenet")
add_model_to_registry(DenseNet161, "imagenet")
add_model_to_registry(DenseNet169, "imagenet")
add_model_to_registry(DenseNet201, "imagenet")



from kimm.blocks.base_block import apply_activation
from kimm.blocks.base_block import apply_conv2d_block
from kimm.blocks.base_block import apply_se_block
from kimm.blocks.depthwise_separation_block import (
    apply_depthwise_separation_block,
)
from kimm.blocks.inverted_residual_block import apply_inverted_residual_block
from kimm.blocks.transformer_block import apply_mlp_block
from kimm.blocks.transformer_block import apply_transformer_block



import math
import typing

import keras
from keras import layers

from kimm.blocks import apply_conv2d_block
from kimm.blocks import apply_depthwise_separation_block
from kimm.blocks import apply_inverted_residual_block
from kimm.models.base_model import BaseModel
from kimm.utils import add_model_to_registry
from kimm.utils import make_divisible

DEFAULT_SMALL_CONFIG = [
    # type, repeat, kernel_size, strides, expansion_ratio, channels, se_ratio,
    # activation
    # stage0
    [["ds", 1, 3, 2, 1.0, 16, 0.25, "relu"]],
    # stage1
    [
        ["ir", 1, 3, 2, 4.5, 24, 0.0, "relu"],
        ["ir", 1, 3, 1, 3.67, 24, 0.0, "relu"],
    ],
    # stage2
    [
        ["ir", 1, 5, 2, 4.0, 40, 0.25, "hard_swish"],
        ["ir", 2, 5, 1, 6.0, 40, 0.25, "hard_swish"],
    ],
    # stage3
    [["ir", 2, 5, 1, 3.0, 48, 0.25, "hard_swish"]],
    # stage4
    [["ir", 3, 5, 2, 6.0, 96, 0.25, "hard_swish"]],
    # stage5
    [["cn", 1, 1, 1, 1.0, 576, 0.0, "hard_swish"]],
]
DEFAULT_LARGE_CONFIG = [
    # type, repeat, kernel_size, strides, expansion_ratio, channels, se_ratio,
    # activation
    # stage0
    [["ds", 1, 3, 1, 1.0, 16, 0.0, "relu"]],
    # stage1
    [
        ["ir", 1, 3, 2, 4.0, 24, 0.0, "relu"],
        ["ir", 1, 3, 1, 3.0, 24, 0.0, "relu"],
    ],
    # stage2
    [["ir", 3, 5, 2, 3.0, 40, 0.25, "relu"]],
    # stage3
    [
        ["ir", 1, 3, 2, 6.0, 80, 0.0, "hard_swish"],
        ["ir", 1, 3, 1, 2.5, 80, 0.0, "hard_swish"],
        ["ir", 2, 3, 1, 2.3, 80, 0.0, "hard_swish"],
    ],
    # stage4
    [["ir", 2, 3, 1, 6.0, 112, 0.25, "hard_swish"]],
    # stage5
    [["ir", 3, 5, 2, 6.0, 160, 0.25, "hard_swish"]],
    # stage6
    [["cn", 1, 1, 1, 1.0, 960, 0.0, "hard_swish"]],
]
DEFAULT_LCNET_CONFIG = [
    # type, repeat, kernel_size, strides, expansion_ratio, channels, se_ratio,
    # activation
    # stage0
    [["dsa", 1, 3, 1, 1.0, 32, 0.0, "hard_swish"]],
    # stage1
    [["dsa", 2, 3, 2, 1.0, 64, 0.0, "hard_swish"]],
    # stage2
    [["dsa", 2, 3, 2, 1.0, 128, 0.0, "hard_swish"]],
    # stage3
    [
        ["dsa", 1, 3, 2, 1.0, 256, 0.0, "hard_swish"],
        ["dsa", 1, 5, 1, 1.0, 256, 0.0, "hard_swish"],
    ],
    # stage4
    [["dsa", 4, 5, 1, 1.0, 256, 0.0, "hard_swish"]],
    # stage5
    [["dsa", 2, 5, 2, 1.0, 512, 0.25, "hard_swish"]],
]


@keras.saving.register_keras_serializable(package="kimm")
class MobileNetV3(BaseModel):
    def __init__(
        self,
        width: float = 1.0,
        depth: float = 1.0,
        fix_stem_and_head_channels: bool = False,
        config: typing.Literal["small", "large", "lcnet"] = "large",
        minimal: bool = False,
        **kwargs,
    ):
        kwargs["weights_url"] = self.get_weights_url(kwargs["weights"])

        _available_configs = ["small", "large", "lcnet"]
        if config == "small":
            _config = DEFAULT_SMALL_CONFIG
            conv_head_channels = 1024
        elif config == "large":
            _config = DEFAULT_LARGE_CONFIG
            conv_head_channels = 1280
        elif config == "lcnet":
            _config = DEFAULT_LCNET_CONFIG
            conv_head_channels = 1280
        else:
            raise ValueError(
                f"config must be one of {_available_configs} using string. "
                f"Received: config={config}"
            )
        if minimal:
            force_activation = "relu"
            force_kernel_size = 3
            no_se = True
        else:
            force_activation = None
            force_kernel_size = None
            no_se = False
        # TF default config
        bn_epsilon = kwargs.pop("bn_epsilon", 1e-5)
        padding = kwargs.pop("padding", None)

        input_tensor = kwargs.pop("input_tensor", None)
        self.set_properties(kwargs)
        inputs = self.determine_input_tensor(
            input_tensor,
            self._input_shape,
            self._default_size,
        )
        x = inputs

        x = self.build_preprocessing(x, "imagenet")

        # Prepare feature extraction
        features = {}

        # stem
        stem_channel = (
            16 if fix_stem_and_head_channels else make_divisible(16 * width)
        )
        x = apply_conv2d_block(
            x,
            stem_channel,
            3,
            2,
            activation=force_activation or "hard_swish",
            bn_epsilon=bn_epsilon,
            padding=padding,
            name="conv_stem",
        )
        features["STEM_S2"] = x

        # blocks
        current_stride = 2
        for current_stage_idx, cfg in enumerate(_config):
            for current_block_idx, sub_cfg in enumerate(cfg):
                block_type, r, k, s, e, c, se, act = sub_cfg

                # override default config
                if force_activation is not None:
                    act = force_activation
                if force_kernel_size is not None:
                    k = force_kernel_size if k > force_kernel_size else k
                if no_se:
                    se = 0.0

                c = make_divisible(c * width)
                # no depth multiplier at first and last block
                if current_block_idx not in (0, len(_config) - 1):
                    r = int(math.ceil(r * depth))
                for current_layer_idx in range(r):
                    s = s if current_layer_idx == 0 else 1
                    _kwargs = {
                        "bn_epsilon": bn_epsilon,
                        "padding": padding,
                        "name": (
                            f"blocks_{current_stage_idx}_"
                            f"{current_block_idx + current_layer_idx}"
                        ),
                    }
                    if block_type in ("ds", "dsa"):
                        x = apply_depthwise_separation_block(
                            x,
                            c,
                            k,
                            1,
                            s,
                            se,
                            act,
                            se_activation="relu",
                            se_gate_activation="hard_sigmoid",
                            se_make_divisible_number=8,
                            pw_activation=act if block_type == "dsa" else None,
                            skip=False if block_type == "dsa" else True,
                            **_kwargs,
                        )
                    elif block_type == "ir":
                        x = apply_inverted_residual_block(
                            x,
                            c,
                            k,
                            1,
                            1,
                            s,
                            e,
                            se,
                            act,
                            se_activation="relu",
                            se_gate_activation="hard_sigmoid",
                            se_make_divisible_number=8,
                            **_kwargs,
                        )
                    elif block_type == "cn":
                        x = apply_conv2d_block(
                            x, c, k, s, activation=act, **_kwargs
                        )
                    current_stride *= s
            features[f"BLOCK{current_stage_idx}_S{current_stride}"] = x

        # Head
        if self._include_top:
            if fix_stem_and_head_channels:
                conv_head_channels = conv_head_channels
            else:
                conv_head_channels = max(
                    conv_head_channels,
                    make_divisible(conv_head_channels * width),
                )
            head_activation = force_activation or "hard_swish"
            x = self.build_top(
                x,
                self._classes,
                self._classifier_activation,
                self._dropout_rate,
                conv_head_channels=conv_head_channels,
                head_activation=head_activation,
            )
        else:
            if self._pooling == "avg":
                x = layers.GlobalAveragePooling2D(name="avg_pool")(x)
            elif self._pooling == "max":
                x = layers.GlobalMaxPooling2D(name="max_pool")(x)

        super().__init__(inputs=inputs, outputs=x, features=features, **kwargs)

        # All references to `self` below this line
        self.width = width
        self.depth = depth
        self.fix_stem_and_head_channels = fix_stem_and_head_channels
        self.config = config
        self.minimal = minimal

    def build_top(
        self,
        inputs,
        classes,
        classifier_activation,
        dropout_rate,
        conv_head_channels,
        head_activation,
    ):
        x = layers.GlobalAveragePooling2D(name="avg_pool", keepdims=True)(
            inputs
        )
        x = layers.Conv2D(
            conv_head_channels, 1, 1, use_bias=True, name="conv_head"
        )(x)
        x = layers.Activation(head_activation, name="act2")(x)
        x = layers.Flatten()(x)
        x = layers.Dropout(rate=dropout_rate, name="conv_head_dropout")(x)
        x = layers.Dense(
            classes, activation=classifier_activation, name="classifier"
        )(x)
        return x

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "width": self.width,
                "depth": self.depth,
                "fix_stem_and_head_channels": self.fix_stem_and_head_channels,
                "config": self.config,
                "minimal": self.minimal,
            }
        )
        return config

    def fix_config(self, config):
        unused_kwargs = [
            "width",
            "depth",
            "fix_stem_and_head_channels",
            "config",
            "minimal",
        ]
        for k in unused_kwargs:
            config.pop(k, None)
        return config


"""
Model Definition
"""


class MobileNet050V3Small(MobileNetV3):
    available_feature_keys = [
        "STEM_S2",
        *[f"BLOCK{i}_S{j}" for i, j in zip(range(6), [4, 8, 16, 16, 32, 32])],
    ]
    available_weights = [
        (
            "imagenet",
            MobileNetV3.default_origin,
            "mobilenet050v3small_mobilenetv3_small_050.lamb_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        config: typing.Union[str, typing.List] = "small",
        name: str = "MobileNet050V3Small",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            0.5,
            1.0,
            True,
            config,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class MobileNet075V3Small(MobileNetV3):
    available_feature_keys = [
        "STEM_S2",
        *[f"BLOCK{i}_S{j}" for i, j in zip(range(6), [4, 8, 16, 16, 32, 32])],
    ]
    available_weights = [
        (
            "imagenet",
            MobileNetV3.default_origin,
            "mobilenet075v3small_mobilenetv3_small_075.lamb_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        config: typing.Union[str, typing.List] = "small",
        name: str = "MobileNet075V3Small",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            0.75,
            1.0,
            False,
            config,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class MobileNet100V3Small(MobileNetV3):
    available_feature_keys = [
        "STEM_S2",
        *[f"BLOCK{i}_S{j}" for i, j in zip(range(6), [4, 8, 16, 16, 32, 32])],
    ]
    available_weights = [
        (
            "imagenet",
            MobileNetV3.default_origin,
            "mobilenet100v3small_mobilenetv3_small_100.lamb_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        config: typing.Union[str, typing.List] = "small",
        name: str = "MobileNet100V3Small",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            1.0,
            1.0,
            False,
            config,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class MobileNet100V3SmallMinimal(MobileNetV3):
    available_feature_keys = [
        "STEM_S2",
        *[f"BLOCK{i}_S{j}" for i, j in zip(range(6), [4, 8, 16, 16, 32, 32])],
    ]
    available_weights = [
        (
            "imagenet",
            MobileNetV3.default_origin,
            (
                "mobilenet100v3smallminimal_"
                "tf_mobilenetv3_small_minimal_100.in1k.keras"
            ),
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        config: typing.Union[str, typing.List] = "small",
        name: str = "MobileNet100V3SmallMinimal",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        # default to TF configuration (bn_epsilon=1e-3 and padding="same")
        super().__init__(
            1.0,
            1.0,
            False,
            config,
            True,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            bn_epsilon=1e-3,
            padding="same",
            **kwargs,
        )


class MobileNet100V3Large(MobileNetV3):
    available_feature_keys = [
        "STEM_S2",
        *[
            f"BLOCK{i}_S{j}"
            for i, j in zip(range(7), [2, 4, 8, 16, 16, 32, 32])
        ],
    ]
    available_weights = [
        (
            "imagenet",
            MobileNetV3.default_origin,
            (
                "mobilenet100v3large_"
                "mobilenetv3_large_100.miil_in21k_ft_in1k.keras"
            ),
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        config: typing.Union[str, typing.List] = "large",
        name: str = "MobileNet100V3Large",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            1.0,
            1.0,
            False,
            config,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )

    def build_preprocessing(self, inputs, mode="imagenet"):
        if (
            self._weights_url is not None
            and "miil_in21k_ft_in1k" in self._weights_url
        ):
            """`miil_in21k_ft_in1k` needs `0_1`"""
            return super().build_preprocessing(inputs, "0_1")
        else:
            return super().build_preprocessing(inputs, mode)


class MobileNet100V3LargeMinimal(MobileNetV3):
    available_feature_keys = [
        "STEM_S2",
        *[
            f"BLOCK{i}_S{j}"
            for i, j in zip(range(7), [2, 4, 8, 16, 16, 32, 32])
        ],
    ]
    available_weights = [
        (
            "imagenet",
            MobileNetV3.default_origin,
            (
                "mobilenet100v3largeminimal_"
                "tf_mobilenetv3_large_minimal_100.in1k.keras"
            ),
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        config: typing.Union[str, typing.List] = "large",
        name: str = "MobileNet100V3LargeMinimal",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        # default to TF configuration (bn_epsilon=1e-3 and padding="same")
        super().__init__(
            1.0,
            1.0,
            False,
            config,
            True,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            bn_epsilon=1e-3,
            padding="same",
            **kwargs,
        )


class LCNet035(MobileNetV3):
    available_feature_keys = [
        "STEM_S2",
        *[f"BLOCK{i}_S{j}" for i, j in zip(range(6), [2, 4, 8, 16, 16, 32])],
    ]
    available_weights = []

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = None,
        config: typing.Union[str, typing.List] = "lcnet",
        name: str = "LCNet035",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        # default to TF configuration (bn_epsilon=1e-3 and padding="same")
        super().__init__(
            0.35,
            1.0,
            False,
            config,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class LCNet050(MobileNetV3):
    available_feature_keys = [
        "STEM_S2",
        *[f"BLOCK{i}_S{j}" for i, j in zip(range(6), [2, 4, 8, 16, 16, 32])],
    ]
    available_weights = [
        (
            "imagenet",
            MobileNetV3.default_origin,
            "lcnet050_lcnet_050.ra2_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",  # TODO: imagenet
        config: typing.Union[str, typing.List] = "lcnet",
        name: str = "LCNet050",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        # default to TF configuration (bn_epsilon=1e-3 and padding="same")
        super().__init__(
            0.5,
            1.0,
            False,
            config,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class LCNet075(MobileNetV3):
    available_feature_keys = [
        "STEM_S2",
        *[f"BLOCK{i}_S{j}" for i, j in zip(range(6), [2, 4, 8, 16, 16, 32])],
    ]
    available_weights = [
        (
            "imagenet",
            MobileNetV3.default_origin,
            "lcnet075_lcnet_075.ra2_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        config: typing.Union[str, typing.List] = "lcnet",
        name: str = "LCNet075",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        # default to TF configuration (bn_epsilon=1e-3 and padding="same")
        super().__init__(
            0.75,
            1.0,
            False,
            config,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class LCNet100(MobileNetV3):
    available_feature_keys = [
        "STEM_S2",
        *[f"BLOCK{i}_S{j}" for i, j in zip(range(6), [2, 4, 8, 16, 16, 32])],
    ]
    available_weights = [
        (
            "imagenet",
            MobileNetV3.default_origin,
            "lcnet100_lcnet_100.ra2_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        config: typing.Union[str, typing.List] = "lcnet",
        name: str = "LCNet100",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        # default to TF configuration (bn_epsilon=1e-3 and padding="same")
        super().__init__(
            1.0,
            1.0,
            False,
            config,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class LCNet150(MobileNetV3):
    available_feature_keys = [
        "STEM_S2",
        *[
            f"BLOCK{i}_S{j}"
            for i, j in zip(range(7), [2, 4, 8, 16, 16, 32, 32])
        ],
    ]
    available_weights = []

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = None,
        config: typing.Union[str, typing.List] = "lcnet",
        name: str = "LCNet150",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        # default to TF configuration (bn_epsilon=1e-3 and padding="same")
        super().__init__(
            1.5,
            1.0,
            False,
            config,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


add_model_to_registry(MobileNet050V3Small, "imagenet")
add_model_to_registry(MobileNet075V3Small, "imagenet")
add_model_to_registry(MobileNet100V3Small, "imagenet")
add_model_to_registry(MobileNet100V3SmallMinimal, "imagenet")
add_model_to_registry(MobileNet100V3Large, "imagenet")
add_model_to_registry(MobileNet100V3LargeMinimal, "imagenet")
add_model_to_registry(LCNet035)
add_model_to_registry(LCNet050, "imagenet")
add_model_to_registry(LCNet075, "imagenet")
add_model_to_registry(LCNet100, "imagenet")
add_model_to_registry(LCNet150)



import typing

import keras
from keras import backend
from keras import layers

from kimm.models import BaseModel
from kimm.utils import add_model_to_registry

DEFAULT_VGG11_CONFIG = [
    64,
    "M",
    128,
    "M",
    256,
    256,
    "M",
    512,
    512,
    "M",
    512,
    512,
    "M",
]
DEFAULT_VGG13_CONFIG = [
    64,
    64,
    "M",
    128,
    128,
    "M",
    256,
    256,
    "M",
    512,
    512,
    "M",
    512,
    512,
    "M",
]
DEFAULT_VGG16_CONFIG = [
    64,
    64,
    "M",
    128,
    128,
    "M",
    256,
    256,
    256,
    "M",
    512,
    512,
    512,
    "M",
    512,
    512,
    512,
    "M",
]
DEFAULT_VGG19_CONFIG = [
    64,
    64,
    "M",
    128,
    128,
    "M",
    256,
    256,
    256,
    256,
    "M",
    512,
    512,
    512,
    512,
    "M",
    512,
    512,
    512,
    512,
    "M",
]


def apply_conv_mlp_layer(
    inputs,
    output_channels,
    kernel_size,
    mlp_ratio=1.0,
    dropout_rate=0.2,
    name="conv_mlp_layer",
):
    mid_channels = int(output_channels * mlp_ratio)

    x = inputs
    x = layers.Conv2D(
        mid_channels, kernel_size, 1, use_bias=True, name=f"{name}_fc1conv2d"
    )(x)
    x = layers.ReLU()(x)
    x = layers.Dropout(dropout_rate, name=f"{name}_drop")(x)
    x = layers.Conv2D(
        output_channels, 1, 1, use_bias=True, name=f"{name}_fc2conv2d"
    )(x)
    x = layers.ReLU()(x)
    return x


@keras.saving.register_keras_serializable(package="kimm")
class VGG(BaseModel):
    available_feature_keys = [
        *[f"BLOCK{i}_S{j}" for i, j in zip(range(6), [1, 2, 4, 8, 16, 32])],
    ]

    def __init__(self, config: typing.Union[str, typing.List], **kwargs):
        kwargs["weights_url"] = self.get_weights_url(kwargs["weights"])

        _available_configs = ["vgg11", "vgg13", "vgg16", "vgg19"]
        if config == "vgg11":
            _config = DEFAULT_VGG11_CONFIG
        elif config == "vgg13":
            _config = DEFAULT_VGG13_CONFIG
        elif config == "vgg16":
            _config = DEFAULT_VGG16_CONFIG
        elif config == "vgg19":
            _config = DEFAULT_VGG19_CONFIG
        else:
            raise ValueError(
                f"config must be one of {_available_configs} using string. "
                f"Received: config={config}"
            )

        input_tensor = kwargs.pop("input_tensor", None)
        self.set_properties(kwargs)
        channels_axis = (
            -1 if backend.image_data_format() == "channels_last" else -3
        )

        inputs = self.determine_input_tensor(
            input_tensor,
            self._input_shape,
            self._default_size,
        )
        x = inputs

        x = self.build_preprocessing(x, "imagenet")

        # Prepare feature extraction
        features = {}

        # Blocks
        current_stage_idx = 0
        current_block_idx = 0
        current_stride = 1
        for c in _config:
            name = f"features_{current_block_idx}"
            if c == "M":
                features[f"BLOCK{current_stage_idx}_S{current_stride}"] = x
                x = layers.MaxPooling2D(2, 2, name=name)(x)
                current_stride *= 2
                current_stage_idx += 1
                current_block_idx += 1
            else:
                x = layers.Conv2D(
                    c,
                    3,
                    1,
                    padding="same",
                    use_bias=True,
                    name=f"features_{current_block_idx}conv2d",
                )(x)
                x = layers.BatchNormalization(
                    axis=channels_axis,
                    momentum=0.9,
                    epsilon=1e-5,
                    name=f"features_{current_block_idx + 1}",
                )(x)
                x = layers.ReLU(name=f"features_{current_block_idx + 2}")(x)
                current_block_idx += 3

        features[f"BLOCK{current_stage_idx}_S{current_stride}"] = x
        x = apply_conv_mlp_layer(x, 4096, 7, 1.0, 0.0, name="pre_logits")

        # Head
        x = self.build_head(x)

        super().__init__(inputs=inputs, outputs=x, features=features, **kwargs)

        # All references to `self` below this line
        self.config = config

    def get_config(self):
        config = super().get_config()
        config.update({"config": self.config})
        return config

    def fix_config(self, config: typing.Dict):
        unused_kwargs = ["config"]
        for k in unused_kwargs:
            config.pop(k, None)
        return config


"""
Model Definition
"""


class VGG11(VGG):
    available_weights = [
        (
            "imagenet",
            VGG.default_origin,
            "vgg11_vgg11_bn.tv_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "VGG11",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            "vgg11",
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class VGG13(VGG):
    available_weights = [
        (
            "imagenet",
            VGG.default_origin,
            "vgg13_vgg13_bn.tv_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "VGG13",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            "vgg13",
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class VGG16(VGG):
    available_weights = [
        (
            "imagenet",
            VGG.default_origin,
            "vgg16_vgg16_bn.tv_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "VGG16",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            "vgg16",
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class VGG19(VGG):
    available_weights = [
        (
            "imagenet",
            VGG.default_origin,
            "vgg19_vgg19_bn.tv_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "VGG19",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            "vgg19",
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


add_model_to_registry(VGG11, "imagenet")
add_model_to_registry(VGG13, "imagenet")
add_model_to_registry(VGG16, "imagenet")
add_model_to_registry(VGG19, "imagenet")



import pathlib
import typing
import urllib.parse

from keras import KerasTensor
from keras import backend
from keras import layers
from keras import models
from keras import utils
from keras.src.applications import imagenet_utils


class BaseModel(models.Model):
    default_origin = (
        "https://github.com/james77777778/kimm/releases/download/0.1.0/"
    )
    available_feature_keys = []
    available_weights = []

    def __init__(
        self,
        inputs,
        outputs,
        features: typing.Optional[typing.Dict[str, KerasTensor]] = None,
        **kwargs,
    ):
        if not hasattr(self, "_feature_extractor"):
            del features
            super().__init__(inputs=inputs, outputs=outputs, **kwargs)
        else:
            if not hasattr(self, "_feature_keys"):
                raise AttributeError(
                    "`self._feature_keys` must be set when initializing "
                    "BaseModel"
                )
            if self._feature_extractor:
                if features is None:
                    raise ValueError(
                        "`features` must be set when `feature_extractor=True`. "
                        f"Received features={features}"
                    )
                if self._feature_keys is None:
                    self._feature_keys = list(features.keys())
                filtered_features = {}
                for k in self._feature_keys:
                    if k not in features:
                        raise KeyError(
                            f"'{k}' is not a key of `features`. Available keys "
                            f"are: {list(features.keys())}"
                        )
                    filtered_features[k] = features[k]
                # Add outputs
                if backend.is_keras_tensor(outputs):
                    filtered_features["TOP"] = outputs
                super().__init__(
                    inputs=inputs, outputs=filtered_features, **kwargs
                )
            else:
                del features
                super().__init__(inputs=inputs, outputs=outputs, **kwargs)

        if hasattr(self, "_weights_url"):
            self.load_pretrained_weights(self._weights_url)

    def set_properties(
        self, kwargs: typing.Dict[str, typing.Any], default_size: int = 224
    ):
        """Must be called in the initilization of the class.

        This method will add following common properties to the model object:
        - input_shape
        - include_preprocessing
        - include_top
        - pooling
        - dropout_rate
        - classes
        - classifier_activation
        - _weights
        - weights_url
        - default_size
        """
        self._input_shape = kwargs.pop("input_shape", None)
        self._include_preprocessing = kwargs.pop("include_preprocessing", True)
        self._include_top = kwargs.pop("include_top", True)
        self._pooling = kwargs.pop("pooling", None)
        self._dropout_rate = kwargs.pop("dropout_rate", 0.0)
        self._classes = kwargs.pop("classes", 1000)
        self._classifier_activation = kwargs.pop(
            "classifier_activation", "softmax"
        )
        self._weights = kwargs.pop("weights", None)
        self._weights_url = kwargs.pop("weights_url", None)
        self._default_size = kwargs.pop("default_size", default_size)
        # feature extractor
        self._feature_extractor = kwargs.pop("feature_extractor", False)
        self._feature_keys = kwargs.pop("feature_keys", None)

    def determine_input_tensor(
        self,
        input_tensor: typing.Optional[KerasTensor] = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        default_size: int = 224,
        min_size: int = 32,
        require_flatten: bool = False,
        static_shape: bool = False,
    ):
        """Determine the input tensor by the arguments."""
        input_shape = imagenet_utils.obtain_input_shape(
            input_shape,
            default_size=default_size,
            min_size=min_size,
            data_format=backend.image_data_format(),
            require_flatten=require_flatten or static_shape,
            weights=None,
        )

        if input_tensor is None:
            x = layers.Input(shape=input_shape)
        else:
            if not backend.is_keras_tensor(input_tensor):
                x = layers.Input(tensor=input_tensor, shape=input_shape)
            else:
                x = utils.get_source_inputs(input_tensor)
        return x

    def build_preprocessing(
        self,
        inputs,
        mode: typing.Literal["imagenet", "0_1", "-1_1"] = "imagenet",
    ):
        if self._include_preprocessing is False:
            return inputs
        channels_axis = (
            -1 if backend.image_data_format() == "channels_last" else -3
        )
        if mode == "imagenet":
            # [0, 255] to [0, 1] and apply ImageNet mean and variance
            x = layers.Rescaling(scale=1.0 / 255.0)(inputs)
            x = layers.Normalization(
                axis=channels_axis,
                mean=[0.485, 0.456, 0.406],
                variance=[0.229, 0.224, 0.225],
            )(x)
        elif mode == "0_1":
            # [0, 255] to [-1, 1]
            x = layers.Rescaling(scale=1.0 / 255.0)(inputs)
        elif mode == "-1_1":
            # [0, 255] to [-1, 1]
            x = layers.Rescaling(scale=1.0 / 127.5, offset=-1.0)(inputs)
        else:
            raise ValueError(
                "`mode` must be one of ('imagenet', '0_1', '-1_1'). "
                f"Received: mode={mode}"
            )
        return x

    def build_top(
        self,
        inputs,
        classes: int,
        classifier_activation: str,
        dropout_rate: float,
    ):
        x = layers.GlobalAveragePooling2D(name="avg_pool")(inputs)
        x = layers.Dropout(rate=dropout_rate, name="head_dropout")(x)
        x = layers.Dense(
            classes, activation=classifier_activation, name="classifier"
        )(x)
        return x

    def build_head(self, inputs):
        x = inputs
        if self._include_top:
            x = self.build_top(
                x,
                self._classes,
                self._classifier_activation,
                self._dropout_rate,
            )
        else:
            if self._pooling == "avg":
                x = layers.GlobalAveragePooling2D(name="avg_pool")(x)
            elif self._pooling == "max":
                x = layers.GlobalMaxPooling2D(name="max_pool")(x)
        return x

    def load_pretrained_weights(self, weights_url: typing.Optional[str] = None):
        if weights_url is not None:
            result = urllib.parse.urlparse(weights_url)
            file_name = pathlib.Path(result.path).name
            weights_path = utils.get_file(
                file_name, weights_url, cache_subdir="kimm_models"
            )
            self.load_weights(weights_path)

    def get_config(self):
        # Don't chain to super here. The default `get_config()` for functional
        # models is nested and cannot be passed to BaseModel.
        config = {
            # models.Model
            "name": self.name,
            "trainable": self.trainable,
            "input_shape": self.input_shape[1:],
            # common
            "include_preprocessing": self._include_preprocessing,
            "include_top": self._include_top,
            "pooling": self._pooling,
            "dropout_rate": self._dropout_rate,
            "classes": self._classes,
            "classifier_activation": self._classifier_activation,
            "weights": self._weights,
            "weights_url": self._weights_url,
            # feature extractor
            "feature_extractor": self._feature_extractor,
            "feature_keys": self._feature_keys,
        }
        return config

    def fix_config(self, config: typing.Dict):
        return config

    def get_weights_url(self, weights):
        if weights is None:
            return None

        for _weights, _origin, _file_name in self.available_weights:
            if weights == _weights:
                return f"{_origin}/{_file_name}"

        # Failed to find the weights
        _available_weights_name = [
            _weights for _weights, _ in self.available_weights
        ]
        raise ValueError(
            f"Available weights are {_available_weights_name}. "
            f"Received weights={weights}"
        )



import typing

import keras
from keras import backend
from keras import initializers
from keras import layers

from kimm import layers as kimm_layers
from kimm.blocks import apply_mlp_block
from kimm.models.base_model import BaseModel
from kimm.utils import add_model_to_registry


def apply_convnext_block(
    inputs,
    output_channels,
    kernel_size,
    strides,
    mlp_ratio,
    activation="gelu",
    use_conv_mlp=False,
    use_grn=False,
    name="convnext_block",
):
    channels_axis = -1 if backend.image_data_format() == "channels_last" else -3
    input_channels = inputs.shape[channels_axis]

    hidden_channels = int(mlp_ratio * output_channels)
    x = inputs
    shortcut = inputs

    # Padding
    padding = "same"
    if strides > 1:
        padding = "valid"
        x = layers.ZeroPadding2D(
            (kernel_size[0] // 2, kernel_size[1] // 2), name=f"{name}_pad"
        )(x)

    # Depthwise
    x = layers.DepthwiseConv2D(
        kernel_size,
        strides,
        padding=padding,
        use_bias=True,
        name=f"{name}_conv_dw_dwconv2d",
    )(x)
    x = layers.LayerNormalization(
        axis=channels_axis, epsilon=1e-6, name=f"{name}_norm"
    )(x)

    # MLP
    x = apply_mlp_block(
        x,
        hidden_channels,
        output_channels,
        activation,
        use_bias=True,
        use_conv_mlp=use_conv_mlp,
        name=f"{name}_mlp",
    )

    # LayerScale
    x = kimm_layers.LayerScale(
        axis=channels_axis,
        initializer=initializers.Constant(1e-6),
        name=f"{name}_layerscale",
    )(x)

    # Downsample
    if input_channels != output_channels or strides != 1:
        shortcut = layers.AveragePooling2D(
            2, strides=strides, name=f"{name}_pool"
        )(shortcut)
        if input_channels != output_channels:
            shortcut = layers.Conv2D(
                output_channels, 1, 1, use_bias=False, name=f"{name}_conv"
            )(shortcut)

    x = layers.Add()([x, shortcut])
    return x


def apply_convnext_stage(
    inputs,
    depth: int,
    output_channels: int,
    kernel_size: int,
    strides: int,
    activation="gelu",
    use_conv_mlp=False,
    use_grn=False,
    name="convnext_stage",
):
    channels_axis = -1 if backend.image_data_format() == "channels_last" else -3
    input_channels = inputs.shape[channels_axis]

    x = inputs

    # Downsample
    if input_channels != output_channels or strides > 1:
        ds_ks = 2 if strides > 1 else 1
        x = layers.LayerNormalization(
            axis=channels_axis, epsilon=1e-6, name=f"{name}_downsample_0"
        )(x)
        x = layers.Conv2D(
            output_channels,
            ds_ks,
            strides,
            padding="valid",
            use_bias=True,
            name=f"{name}_downsample_1_conv2d",
        )(x)

    for i in range(depth):
        x = apply_convnext_block(
            x,
            output_channels,
            kernel_size,
            1,
            mlp_ratio=4,
            activation=activation,
            use_conv_mlp=use_conv_mlp,
            use_grn=use_grn,
            name=f"{name}_blocks_{i}",
        )
    return x


@keras.saving.register_keras_serializable(package="kimm")
class ConvNeXt(BaseModel):
    available_feature_keys = [
        "STEM_S4",
        *[f"BLOCK{i}_S{2**(i+2)}" for i in range(4)],
    ]

    def __init__(
        self,
        depths: typing.Sequence[int] = [3, 3, 9, 3],
        hidden_channels: typing.Sequence[int] = [96, 192, 384, 768],
        patch_size: int = 4,
        kernel_size: int = 7,
        activation: str = "gelu",
        use_conv_mlp: bool = False,
        **kwargs,
    ):
        kwargs["weights_url"] = self.get_weights_url(kwargs["weights"])
        input_tensor = kwargs.pop("input_tensor", None)
        self.set_properties(kwargs)
        channels_axis = (
            -1 if backend.image_data_format() == "channels_last" else -3
        )

        inputs = self.determine_input_tensor(
            input_tensor,
            self._input_shape,
            self._default_size,
        )
        x = inputs

        x = self.build_preprocessing(x, "imagenet")

        # Prepare feature extraction
        features = {}

        # Stem
        x = layers.Conv2D(
            hidden_channels[0],
            patch_size,
            patch_size,
            use_bias=True,
            name="stem_0_conv2d",
        )(x)
        x = layers.LayerNormalization(
            axis=channels_axis, epsilon=1e-6, name="stem_1"
        )(x)
        features["STEM_S4"] = x

        # Blocks (4 stages)
        current_stride = patch_size
        for i in range(4):
            strides = 2 if i > 0 else 1
            x = apply_convnext_stage(
                x,
                depths[i],
                hidden_channels[i],
                kernel_size,
                strides,
                activation,
                use_conv_mlp,
                use_grn=False,
                name=f"stages_{i}",
            )
            current_stride *= strides
            # Add feature
            features[f"BLOCK{i}_S{current_stride}"] = x

        # Head
        x = self.build_head(x)

        super().__init__(inputs=inputs, outputs=x, features=features, **kwargs)

        # All references to `self` below this line
        self.depths = depths
        self.hidden_channels = hidden_channels
        self.patch_size = patch_size
        self.kernel_size = kernel_size
        self.activation = activation
        self.use_conv_mlp = use_conv_mlp

    def build_top(self, inputs, classes, classifier_activation, dropout_rate):
        x = inputs
        x = layers.GlobalAveragePooling2D(name="avg_pool")(inputs)
        x = layers.LayerNormalization(axis=-1, epsilon=1e-6, name="head_norm")(
            x
        )
        x = layers.Dropout(rate=dropout_rate, name="head_dropout")(x)
        x = layers.Dense(
            classes, activation=classifier_activation, name="classifier"
        )(x)
        return x

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "depths": self.depths,
                "hidden_channels": self.hidden_channels,
                "patch_size": self.patch_size,
                "kernel_size": self.kernel_size,
                "activation": self.activation,
                "use_conv_mlp": self.use_conv_mlp,
            }
        )
        return config

    def fix_config(self, config):
        unused_kwargs = [
            "depths",
            "hidden_channels",
            "patch_size",
            "kernel_size",
            "activation",
            "use_conv_mlp",
        ]
        for k in unused_kwargs:
            config.pop(k, None)
        return config


"""
Model Definition
"""


class ConvNeXtAtto(ConvNeXt):
    available_weights = [
        (
            "imagenet",
            ConvNeXt.default_origin,
            "convnextatto_convnext_atto.d2_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "ConvNeXtAtto",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            (2, 2, 6, 2),
            (40, 80, 160, 320),
            4,
            7,
            "gelu",
            True,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class ConvNeXtFemto(ConvNeXt):
    available_weights = [
        (
            "imagenet",
            ConvNeXt.default_origin,
            "convnextfemto_convnext_femto.d1_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "ConvNeXtFemto",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            (2, 2, 6, 2),
            (48, 96, 192, 384),
            4,
            7,
            "gelu",
            True,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class ConvNeXtPico(ConvNeXt):
    available_weights = [
        (
            "imagenet",
            ConvNeXt.default_origin,
            "convnextpico_convnext_pico.d1_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "ConvNeXtPico",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            (2, 2, 6, 2),
            (64, 128, 256, 512),
            4,
            7,
            "gelu",
            True,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class ConvNeXtNano(ConvNeXt):
    available_weights = [
        (
            "imagenet",
            ConvNeXt.default_origin,
            "convnextnano_convnext_nano.in12k_ft_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "ConvNeXtNano",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            (2, 2, 8, 2),
            (80, 160, 320, 640),
            4,
            7,
            "gelu",
            True,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class ConvNeXtTiny(ConvNeXt):
    available_weights = [
        (
            "imagenet",
            ConvNeXt.default_origin,
            "convnexttiny_convnext_tiny.in12k_ft_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "ConvNeXtTiny",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            (3, 3, 9, 3),
            (96, 192, 384, 768),
            4,
            7,
            "gelu",
            False,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class ConvNeXtSmall(ConvNeXt):
    available_weights = [
        (
            "imagenet",
            ConvNeXt.default_origin,
            "convnextsmall_convnext_small.in12k_ft_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "ConvNeXtSmall",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            (3, 3, 27, 3),
            (96, 192, 384, 768),
            4,
            7,
            "gelu",
            False,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class ConvNeXtBase(ConvNeXt):
    available_weights = [
        (
            "imagenet",
            ConvNeXt.default_origin,
            "convnextbase_convnext_base.fb_in22k_ft_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "ConvNeXtBase",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            (3, 3, 27, 3),
            (128, 256, 512, 1024),
            4,
            7,
            "gelu",
            False,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class ConvNeXtLarge(ConvNeXt):
    available_weights = [
        (
            "imagenet",
            ConvNeXt.default_origin,
            "convnextlarge_convnext_large.fb_in22k_ft_in1k.keras",
        )
    ]

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = "imagenet",
        name: str = "ConvNeXtLarge",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            (3, 3, 27, 3),
            (192, 384, 768, 1536),
            4,
            7,
            "gelu",
            False,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


class ConvNeXtXLarge(ConvNeXt):
    available_weights = []

    def __init__(
        self,
        input_tensor: keras.KerasTensor = None,
        input_shape: typing.Optional[typing.Sequence[int]] = None,
        include_preprocessing: bool = True,
        include_top: bool = True,
        pooling: typing.Optional[str] = None,
        dropout_rate: float = 0.0,
        classes: int = 1000,
        classifier_activation: str = "softmax",
        weights: typing.Optional[str] = None,
        name: str = "ConvNeXtXLarge",
        **kwargs,
    ):
        kwargs = self.fix_config(kwargs)
        super().__init__(
            (3, 3, 27, 3),
            (256, 512, 1024, 2048),
            4,
            7,
            "gelu",
            False,
            input_tensor=input_tensor,
            input_shape=input_shape,
            include_preprocessing=include_preprocessing,
            include_top=include_top,
            pooling=pooling,
            dropout_rate=dropout_rate,
            classes=classes,
            classifier_activation=classifier_activation,
            weights=weights,
            name=name,
            **kwargs,
        )


add_model_to_registry(ConvNeXtAtto, "imagenet")
add_model_to_registry(ConvNeXtFemto, "imagenet")
add_model_to_registry(ConvNeXtPico, "imagenet")
add_model_to_registry(ConvNeXtNano, "imagenet")
add_model_to_registry(ConvNeXtTiny, "imagenet")
add_model_to_registry(ConvNeXtSmall, "imagenet")
add_model_to_registry(ConvNeXtBase, "imagenet")
add_model_to_registry(ConvNeXtLarge, "imagenet")
add_model_to_registry(ConvNeXtXLarge)



from kimm.models.base_model import BaseModel
from kimm.models.convmixer import *  # noqa:F403
from kimm.models.convnext import *  # noqa:F403
from kimm.models.densenet import *  # noqa:F403
from kimm.models.efficientnet import *  # noqa:F403
from kimm.models.ghostnet import *  # noqa:F403
from kimm.models.inception_v3 import *  # noqa:F403
from kimm.models.mobilenet_v2 import *  # noqa:F403
from kimm.models.mobilenet_v3 import *  # noqa:F403
from kimm.models.mobilevit import *  # noqa:F403
from kimm.models.regnet import *  # noqa:F403
from kimm.models.resnet import *  # noqa:F403
from kimm.models.vgg import *  # noqa:F403
from kimm.models.vision_transformer import *  # noqa:F403
from kimm.models.xception import *  # noqa:F403
